import json
import logging


def make_compat_start_v1(config, settings, info):
    return json.dumps(
        {
            "runId": settings._op_id,
            "runName": settings._op_name,
            "projectName": settings.project,
            "metadata": json.dumps(config),
            "systemMetadata": json.dumps(info),
        }
    ).encode()


def make_compat_stop_v1(data, settings):
    return json.dumps(
        {
            "runId": settings._op_id,
            "status": data,
            "statusMetadata": json.dumps(settings.meta),
        }
    ).encode()


def make_compat_meta_v1(meta, settings):
    return json.dumps(
        {
            "runId": settings._op_id,
            # "runName": settings._op_name,
            # "projectName": settings.project,
            "logName": meta,  # TODO: better aggregate
        }
    ).encode()


def make_compat_data_v1(data, timestamp, step):
    line = [
        json.dumps(
            {
                "time": int(timestamp * 1000),  # convert to ms
                "step": int(step),
                "data": data,
            }
        )
    ]
    return ("\n".join(line) + "\n").encode("utf-8")


def make_compat_file_v1(file, timestamp, step):
    batch = []
    for k, fl in file.items():
        for f in fl:
            i = {
                "fileName": f"{f._name}{f._ext}",
                "size": f._size,
                "fileType": f._ext[1:],
                "logName": k,
                "step": step,
            }
            batch.append(i)
    return json.dumps({"files": batch}).encode()


def make_compat_storage_v1(f, fl):
    # workaround for lack of file ident on server side
    for i in fl:
        if next(iter(i.keys())) == f"{f._name}{f._ext}":
            return next(iter(i.values()))
    return None


def make_compat_message_v1(level, message, timestamp, step):
    line = [
        json.dumps(
            {
                "time": int(timestamp * 1000),  # convert to ms
                "message": message,
                "lineNumber": step,
                "logType": "INFO" if level == logging.INFO else "ERROR",
            }
        )
    ]
    return ("\n".join(line) + "\n").encode("utf-8")



from .auth import login, logout
from .file import File, Image
from .init import init
from .sets import Settings, setup
from .sys import System

# TODO: setup preinit
__all__ = (
    "File",
    "Image",
    "System",
    "Settings",
    "init",
    "login",
    "logout",
    "setup",
)

__version__ = "0.0.0"



import importlib.metadata
import logging
import os
import platform
import time

import psutil
from git import Repo

from .sets import Settings
from .util import to_human  # TODO: move to server side

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "System"


class System:
    def __init__(self, settings: Settings) -> None:
        self.settings = settings

        self.uname = platform.uname()._asdict()
        self.timezone = list(time.tzname)

        self.cpu_count = psutil.cpu_count
        self.cpu_freq = [i._asdict() for i in psutil.cpu_freq(percpu=True)]
        self.cpu_freq_min = min([i["min"] for i in self.cpu_freq])
        self.cpu_freq_max = max([i["max"] for i in self.cpu_freq])
        self.svmem = psutil.virtual_memory()._asdict()
        self.sswap = psutil.swap_memory()._asdict()
        self.disk = [i._asdict() for i in psutil.disk_partitions()]
        self.net_if_addrs = {
            i: [
                {k: v for k, v in j._asdict().items() if k != "family"}
                for j in psutil.net_if_addrs()[i]
            ]
            for i in psutil.net_if_addrs()
        }
        self.boot_time = psutil.boot_time()
        self.users = [i._asdict() for i in psutil.users()]

        self.pid = os.getpid()
        self.proc = psutil.Process(pid=self.pid)
        with self.proc.oneshot():  # perf
            self.proc_info = self.proc.as_dict(attrs=["exe", "cmdline"])
            self.proc_child = self.proc.children(recursive=True)
            self.pid_child = [p.pid for p in self.proc_child] + [self.pid]
        if self.settings.mode == "debug":  # privacy guard
            self.environ = self.proc.environ()
            self.requirements = [
                f"{p.metadata['Name']}=={p.version}"
                for p in importlib.metadata.distributions()
            ]

        self.gpu = self.get_gpu()
        self.git = self.get_git()

    def __getattr__(self, name):
        return self.get_psutil(name)

    def get_psutil(self, name):  # handling os specific methods
        if hasattr(psutil, name):
            return getattr(psutil, name)
        else:
            return None

    def get_gpu(self):
        d = {}
        try:  # NVIDIA
            import pynvml

            try:
                pynvml.nvmlInit()
                logger.info(f"{tag}: NVIDIA GPU detected")
                d["nvidia"] = {
                    "count": pynvml.nvmlDeviceGetCount(),
                    "driver": pynvml.nvmlSystemGetDriverVersion(),
                    "devices": [],
                    "handles": [],
                }
                for i in range(d["nvidia"]["count"]):
                    h = pynvml.nvmlDeviceGetHandleByIndex(i)
                    d["nvidia"]["handles"].append(h)
                    d["nvidia"]["devices"].append(
                        {
                            "name": pynvml.nvmlDeviceGetName(h),
                            "memory": {
                                "total": to_human(
                                    pynvml.nvmlDeviceGetMemoryInfo(h).total
                                ),
                            },
                            "temp": pynvml.nvmlDeviceGetTemperature(
                                h, pynvml.NVML_TEMPERATURE_GPU
                            ),
                            "pid": [
                                p.pid
                                for p in (
                                    pynvml.nvmlDeviceGetComputeRunningProcesses(h)
                                    + pynvml.nvmlDeviceGetGraphicsRunningProcesses(h)
                                )
                            ],
                        }
                    )
            except pynvml.NVMLError_LibraryNotFound:
                logger.debug(f"{tag}: NVIDIA: driver not found")
            except Exception as e:
                logger.error("%s: NVIDIA: error: %s", tag, e)
        except ImportError:
            logger.debug(f"{tag}: NVIDIA: pynvml not found")
        return d

    def get_git(self):
        d = {}
        try:
            repo = Repo(
                f"{self.settings.work_dir()}" or os.getcwd(),
                search_parent_directories=True,
            )
            try:
                url = repo.remotes["origin"].url
            except IndexError:
                url = None
            d = {
                "url": url,
                # "name": repo.config_reader().get_value("user", "name"),
                "email": repo.config_reader().get_value("user", "email"),
                "root": repo.git.rev_parse("--show-toplevel"),
                "dirty": repo.is_dirty(),
                "branch": repo.head.ref.name,
                "commit": repo.head.commit.hexsha,
            }
        except Exception as e:
            logger.debug("%s: git: repository not detected: %s", tag, e)
        return d

    def info(self):
        d = {
            "process": {
                **self.proc_info,
                "pid": self.pid,
            },
            "platform": self.uname,
            "timezone": self.timezone,
            "cpu": {
                "physical": self.cpu_count(logical=False),
                "virtual": self.cpu_count(logical=True),
                "freq": {
                    "min": self.cpu_freq_min,
                    "max": self.cpu_freq_max,
                },
            },
            "memory": {
                "virt": to_human(self.svmem["total"]),
                "swap": to_human(self.sswap["total"]),
            },
            "boot_time": self.boot_time,
        }
        if self.gpu:
            d["gpu"] = {}
            if self.gpu.get("nvidia"):
                d["gpu"]["nvidia"] = {
                    k: v for k, v in self.gpu["nvidia"].items() if k != "handles"
                }
        if self.git:
            d["git"] = self.git
        if self.settings.mode == "debug":
            d["process"]["environ"] = self.environ
            d = {
                **d,
                "disk": self.disk,
                "network": self.net_if_addrs,
                "users": self.users,
                "requirements": self.requirements,
            }
        return d

    def monitor(self):
        p = self.settings.x_sys_label
        d = {
            **{
                f"{p}cpu/pct/{i}": v
                for i, v in enumerate(psutil.cpu_percent(percpu=True))
            },
            **{
                f"{p}mem/{k}": v
                for k, v in psutil.virtual_memory()._asdict().items()
                if k in ("active", "used")
            },
            **{
                f"{p}disk/{k}": v
                for k, v in psutil.disk_usage(self.settings.work_dir())
                ._asdict()
                .items()
                if k in ("used")
            },
            **{
                f"{p}net/{k}": v
                for k, v in psutil.net_io_counters()._asdict().items()
                if k.startswith("bytes")
            },
        }
        if self.gpu:
            if self.gpu.get("nvidia"):
                import pynvml

                for h in self.gpu["nvidia"]["handles"]:
                    dev = (
                        str(pynvml.nvmlDeviceGetName(h))[2:-1].lower().replace(" ", "_")
                    )
                    d[f"{p}gpu/nvda/{dev}/pct"] = pynvml.nvmlDeviceGetUtilizationRates(
                        h
                    ).gpu
                    d[f"{p}gpu/nvda/{dev}/mem/pct"] = (
                        pynvml.nvmlDeviceGetUtilizationRates(h).memory
                    )
                    d[f"{p}gpu/nvda/{dev}/power"] = pynvml.nvmlDeviceGetPowerUsage(h)
        return d

    def monitor_human(self):
        d = {
            "cpu": {
                "percent": psutil.cpu_percent(percpu=True),
                "freq": [i.current for i in psutil.cpu_freq(percpu=True)],
            },
            "memory": {
                "virt": {
                    k: to_human(v)
                    for k, v in psutil.virtual_memory()._asdict().items()
                    if k != "percent"
                },
            },
            "disk": {
                "out": to_human(psutil.disk_io_counters().read_bytes),
                "in": to_human(psutil.disk_io_counters().write_bytes),
                "usage": {
                    k: to_human(v)
                    for k, v in psutil.disk_usage(self.settings.work_dir())
                    ._asdict()
                    .items()
                    if k != "percent"
                },
            },
            "network": {
                "out": to_human(psutil.net_io_counters().bytes_sent),
                "in": to_human(psutil.net_io_counters().bytes_recv),
            },
        }
        with self.proc.oneshot():  # perf
            d["process"] = {
                **self.proc.as_dict(
                    attrs=["status", "cpu_percent", "memory_percent", "num_threads"]
                ),
                "memory": to_human(self.proc.memory_info().rss),
            }
        if self.gpu:
            d["gpu"] = {}
            if self.gpu.get("nvidia"):
                import pynvml

                d["gpu"]["nvidia"] = {}
                d["gpu"]["nvidia"]["devices"] = [
                    {
                        "name": pynvml.nvmlDeviceGetName(h),
                        "temp": pynvml.nvmlDeviceGetTemperature(
                            h, pynvml.NVML_TEMPERATURE_GPU
                        ),
                        "gpu_percent": pynvml.nvmlDeviceGetUtilizationRates(h).gpu,
                        "memory": {
                            "percent": pynvml.nvmlDeviceGetUtilizationRates(h).memory,
                            "used": to_human(pynvml.nvmlDeviceGetMemoryInfo(h).used),
                            "total": to_human(pynvml.nvmlDeviceGetMemoryInfo(h).total),
                        },
                        "power": {
                            "usage": pynvml.nvmlDeviceGetPowerUsage(h),
                            "limit": pynvml.nvmlDeviceGetEnforcedPowerLimit(h),
                        },
                        "pid": [
                            p.pid
                            for p in (
                                pynvml.nvmlDeviceGetComputeRunningProcesses(h)
                                + pynvml.nvmlDeviceGetGraphicsRunningProcesses(h)
                            )
                        ],
                    }
                    for h in self.gpu["nvidia"]["handles"]
                ]
        if self.settings.mode == "debug":
            d["memory"]["swap"] = {
                k: to_human(v)
                for k, v in psutil.swap_memory()._asdict().items()
                if k != "percent"
            }
        # exit(1)
        return d



import logging
import queue
import threading
import time

import httpx
import keyring

from .api import (
    make_compat_data_v1,
    make_compat_file_v1,
    make_compat_meta_v1,
    make_compat_start_v1,
    make_compat_stop_v1,
    make_compat_storage_v1,
)
from .auth import login
from .sets import Settings
from .util import print_url

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "Interface"


class ServerInterface:
    def __init__(self, config: dict, settings: Settings) -> None:
        login()
        self.config = config
        self.settings = settings
        self.settings.auth = keyring.get_password(
            f"{self.settings.tag}", f"{self.settings.tag}"
        )

        # self.url_view = f"{self.settings.url_view}/{self.settings.user}/{self.settings.project}/{self.settings._op_id}"
        self.headers = {
            "Authorization": f"Bearer {self.settings.auth}",
            "Content-Type": "application/json",
            "User-Agent": f"{self.settings.tag}",
            "X-Run-Id": f"{self.settings._op_id}",
            "X-Run-Name": f"{self.settings._op_name}",
            "X-Project-Name": f"{self.settings.project}",
        }
        self.headers_data = self.headers.copy()
        self.headers_data.update({"Content-Type": "application/x-ndjson"})

        self.client = httpx.Client(
            # http2=True,
            verify=True if not self.settings.insecure_disable_ssl else False,
            proxy=self.settings.http_proxy or self.settings.https_proxy or None,
            limits=httpx.Limits(
                max_keepalive_connections=self.settings.x_file_stream_max_conn,
                max_connections=self.settings.x_file_stream_max_conn,
            ),
            timeout=httpx.Timeout(
                self.settings.x_file_stream_timeout_seconds,
                # connect=settings.x_file_stream_timeout_seconds,
            ),
        )

        self.client_storage = httpx.Client(
            # http1=False, # TODO: set http2
            verify=True if not self.settings.insecure_disable_ssl else False,
            proxy=self.settings.http_proxy or self.settings.https_proxy or None,
        )

        self._stop_event = threading.Event()

        self._queue_data = queue.Queue()
        self._buffer_data = None
        self._thread_data = None
        self._thread_file = None
        self._thread_storage = None

        self._queue_message = self.settings.message
        self._buffer_message = None
        self._thread_message = None

    def start(self) -> None:
        r = self._post_v1(
            self.settings.url_start,
            self.headers,
            make_compat_start_v1(self.config, self.settings, self.settings.system.info()),
            client=self.client,
        )
        logger.info(
            f"{tag}: started run {r.json()['runId']}"
        )  # TODO: send a proper response
        self.settings.url_view = r.json()["url"]
        logger.info(f"{tag}: find live updates at {print_url(self.settings.url_view)}")
        self._update_meta(list(self.settings.system.monitor().keys()))

        if self._thread_data is None:
            self._thread_data = threading.Thread(
                target=self._worker_publish,
                args=(
                    self.settings.url_data,
                    self.headers_data,
                    self._queue_data,
                    self._buffer_data,
                    self._stop_event.is_set,
                    "data",
                ),
                daemon=True,
            )
            self._thread_data.start()
        if self._thread_message is None:
            self._thread_message = threading.Thread(
                target=self._worker_publish,
                args=(
                    self.settings.url_message,
                    self.headers,
                    self._queue_message,
                    self._buffer_message,
                    self._stop_event.is_set,
                    "message" if self.settings.mode == "debug" else None,
                ),
                daemon=True,
            )
            self._thread_message.start()

    def publish(
        self,
        data: dict[str, any] | None = None,
        file: None = None,
        timestamp: int | None = None,
        step: int | None = None,
    ) -> None:
        if data:
            self._queue_data.put(
                make_compat_data_v1(data, timestamp, step), block=False
            )
        if file:
            self._thread_file = threading.Thread(
                target=self._worker_file,
                args=(file, make_compat_file_v1(file, timestamp, step)),
                daemon=True,
            )  # TODO: batching
            self._thread_file.start()

    def stop(self) -> None:
        self._stop_event.set()
        for t in [
            self._thread_data,
            self._thread_file,
            self._thread_storage,
            self._thread_message,
        ]:
            if t is not None:
                t.join(timeout=None)
                t = None
        r = self._post_v1(
            self.settings.url_stop,
            self.headers,
            make_compat_stop_v1("COMPLETED", self.settings),
            client=self.client,
        )
        logger.info(f"{tag}: find uploaded data at {print_url(self.settings.url_view)}")

    def _update_meta(self, meta):
        r = self._post_v1(
            self.settings.url_meta,
            self.headers,
            make_compat_meta_v1(meta, self.settings),
            client=self.client,
        )

    def _worker_publish(self, e, h, q, b, stop, name=None):
        while not (q.empty() and stop()):  # terminates only when both conditions met
            if q.empty():
                time.sleep(self.settings.x_internal_check_process) # debounce
            else:
                _ = self._post_v1(
                    e,
                    h,
                    q,
                    b,
                    client=self.client,
                    name=name,
                )

    def _worker_storage(self, f):
        _ = self._put_v1(
            f._url,
            {
                "Content-Type": f._type,  # "application/octet-stream"
            },
            open(f._path, "rb"),
            client=self.client_storage,
        )

    def _worker_file(self, file, q):
        r = self._post_v1(
            self.settings.url_file,
            self.headers,
            q,
            client=self.client,
        )
        try:
            d = r.json()
            logger.debug(f"{tag}: file api responded {len(d)} key(s)")
            for k, fel in file.items():
                for f in fel:
                    f._url = make_compat_storage_v1(f, d[k])
                    if not f._url:
                        logger.critical(f"{tag}: file api did not provide storage url")
                    else:
                        self._thread_storage = threading.Thread(
                            target=self._worker_storage, args=(f,), daemon=True
                        )
                        self._thread_storage.start()
        except Exception as e:
            logger.critical("%s: failed to send files to %s: %s", tag, self.settings.url_file, e)

    def _queue_iter(self, q, b):
        s = time.time()
        while (
            len(b) < self.settings.x_file_stream_max_size
            and (time.time() - s) < self.settings.x_file_stream_transmit_interval
        ):
            try:
                v = q.get(timeout=self.settings.x_internal_check_process)
                b.append(v)
                yield v
            except queue.Empty:
                break

    def _put_v1(self, url, headers, content, client=None, retry=0):
        try:
            r = client.put(
                url,
                content=content,
                headers=headers,
            )
            if r.status_code in [200, 201]:
                logger.debug(f"{tag}: successfully put one item in storage")
                return r
            else:
                logger.error(
                    f"{tag}: server responded error {r.status_code} during PUT to {url}: {r.text}"
                )
        except Exception as e:
            logger.error("%s: no response received during PUT to %s: %s", tag, url, e)
        retry += 1
        self._put_v1(
            url, headers, content, client=client, retry=retry
        ) if retry < self.settings.x_file_stream_retry_max else logger.critical(
            f"{tag}: failed to put item in storage after {retry} retries to {url}"
        )

    def _post_v1(self, url, headers, q, b=[], client=None, name=None, retry=0):
        b = []
        try:
            s = time.time()
            content = self._queue_iter(q, b) if isinstance(q, queue.Queue) else q
            r = client.post(
                url,
                content=content,  # iter(q.get, None),
                headers=headers,
            )
            if r.status_code in [200, 201]:
                if name is not None and isinstance(q, queue.Queue):
                    logger.debug(
                        f"{tag}: {name}: sent {len(b)} line(s) at {len(b) / (time.time() - s):.2f} lines/s to {url}"
                    )
                return r
            else:
                if name is not None:
                    logger.error(
                        f"{tag}: {name}: server responded error {r.status_code} for {len(b)} line(s) during POST: {r.text}"
                    )
        except Exception as e:
            logger.error("%s: no response received during POST to %s: %s", tag, url, e)

        retry += 1
        if retry < self.settings.x_file_stream_retry_max:
            logger.warning(
                f"{tag}: retry {retry}/{self.settings.x_file_stream_retry_max} - server responded error {r.status_code} for {len(b)} line(s) during POST to {url}: {r.text}"
            )
            time.sleep(
                min(
                    self.settings.x_file_stream_retry_wait_min_seconds * (2**retry),
                    self.settings.x_file_stream_retry_wait_max_seconds,
                )
            )
            for i in b:  # if isinstance(q, queue.Queue)
                q.put(i, block=False)
            return self._post_v1(
                url, headers, q, b, client=client, retry=retry
            )  # kwargs
        else:
            if name is not None:
                logger.critical(
                    f"{tag}: {name}: failed to send {len(b)} line(s) after {retry} retries"
                )
            return None



import atexit
import logging
import multiprocessing
import queue
import threading
import time
from collections.abc import Mapping

from .file import File, Image
from .iface import ServerInterface
from .log import teardown_logger
from .sets import Settings
from .store import DataStore
from .util import dict_to_json

logger = logging.getLogger(f"{__name__.split('.')[0]}")
TAG = "Logging"


class OpsMonitor:
    def __init__(self, op) -> None:
        self.op = op
        self._stop_event = threading.Event()
        self._thread = None
        self._thread_monitor = None

    def start(self) -> None:
        if self._thread is None:
            self._thread = threading.Thread(
                target=self.op._worker, args=(self._stop_event.is_set,), daemon=True
            )
            self._thread.start()
        if self._thread_monitor is None:
            self._thread_monitor = threading.Thread(
                target=self._worker_monitor,
                args=(self._stop_event.is_set,),
                daemon=True,
            )
            self._thread_monitor.start()

    def stop(self) -> None:
        self._stop_event.set()
        for t in [self._thread, self._thread_monitor]:
            if t is not None:
                t.join()  # timeout=self.op.settings.x_sys_sampling_interval
                t = None

    def _worker_monitor(self, stop):
        while not stop():
            self.op._iface.publish(
                data=self.op.settings.system.monitor(),
                file=None,
                timestamp=int(time.time()),
                step=self.op._step,
            ) if self.op._iface else None
            time.sleep(self.op.settings.x_sys_sampling_interval)


class Ops:
    def __init__(self, config, settings) -> None:
        self.config = config
        self.settings = settings
        self._monitor = OpsMonitor(op=self)

        self._store = (
            DataStore(config=config, settings=settings)
            if not settings.disable_store
            else None
        )
        self._iface = (
            ServerInterface(config=config, settings=settings)
            if not settings.disable_iface
            else None
        )
        self._step = 0
        self._queue = queue.Queue()
        atexit.register(self.finish)

    def start(self) -> None:
        self._iface.start() if self._iface else None
        self._monitor.start()
        logger.debug(f"{TAG}: started")

    def log(
        self, data: dict[str, any], step: int | None = None, commit: bool | None = None
    ) -> None:
        """Log run data"""
        if self.settings.mode == "perf":
            self._queue.put((data, step), block=False)
        else:  # bypass queue
            self._log(data=data, step=step)

    def finish(self, exit_code: int | None = None) -> None:
        """Finish logging"""
        self._finish(exit_code=exit_code)
        while not self._queue.empty():
            pass
        self._store.stop() if self._store else None
        self._iface.stop() if self._iface else None  # fixed order
        logger.debug(f"{TAG}: finished")
        teardown_logger(logger, console=logging.getLogger("console"))

    def _worker(self, stop) -> None:
        while not stop() or not self._queue.empty():
            try:
                # if queue seems empty, wait for x_internal_check_process before it considers it empty to save compute
                self._log(
                    *self._queue.get(
                        block=True, timeout=self.settings.x_internal_check_process
                    )
                )
            except queue.Empty:
                continue
            except Exception as e:
                time.sleep(self.settings.x_internal_check_process)  # debounce
                logger.critical("%s: failed: %s", TAG, e)
                raise e

    def _log(self, data, step) -> None:
        if not isinstance(data, Mapping):
            e = ValueError(
                f"Data logged must be of dictionary type; received {type(data).__name__} intsead"
            )
            logger.critical("%s: failed: %s", TAG, e)
            raise e
        if any(not isinstance(k, str) for k in data.keys()):
            e = ValueError("Data logged must have keys of string type")
            logger.critical("%s: failed: %s", TAG, e)
            raise e

        t = int(time.time())
        if step is not None:
            if step > self._step:
                self._step = step
        else:
            self._step += 1

        # data = data.copy()  # TODO: check mutability
        d, f, m = {}, {}, []
        for k, v in data.items():
            if k not in self.settings.meta:
                m.append(k)
                self.settings.meta.append(k)
                # d[f"{self.settings.x_meta_label}{k}"] = 0
                logger.debug(f"{TAG}: added {k} at step {self._step}")
            if isinstance(v, list):
                for e in v:
                    d, f = self._op(d, f, k, e)
            else:
                d, f = self._op(d, f, k, v)

        # d = dict_to_json(d)  # TODO: add serialisation
        self._store.insert(
            data=d, file=f, timestamp=t, step=self._step
        ) if self._store else None
        self._iface.publish(
            data=d, file=f, timestamp=t, step=self._step
        ) if self._iface else None
        self._iface._update_meta(m) if m and self._iface else None

    def _op(self, d, f, k, v) -> None:
        if isinstance(v, File):
            if isinstance(v, Image):
                v.load(self.settings.work_dir())
            # TODO: add step to serialise data for files
            v._mkcopy(self.settings.work_dir())  # key independent
            # d[k] = int(v._id, 16)
            if k not in f:
                f[k] = [v]
            else:
                f[k].append(v)
        elif isinstance(v, (int, float)):
            d[k] = v
        else:
            pass  # raise not supported error
        return d, f

    def _finish(self, exit_code) -> None:
        self._monitor.stop()



import logging
import os
import queue
import sys

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "Setup"


class Settings:
    tag: str = f"{__name__.split('.')[0]}"
    dir: str = str(os.path.abspath(os.getcwd()))

    auth: str = None
    project: str = "default"
    mode: str = "perf"  # noop | debug | perf
    system: dict[str, any] = {}
    meta: list = None
    message: queue.Queue = queue.Queue()
    disable_store: bool = True  # TODO: make false
    disable_iface: bool = False
    disable_logger: bool = False  # disable file-based logging

    _op_name: str = None
    _op_id: str = None

    store_db: str = "store.db"
    store_table_data: str = "data"
    store_table_file: str = "file"
    store_max_size: int = 2**14
    store_aggregate_interval: float = 2 ** (-1)

    http_proxy: str = None
    https_proxy: str = None
    insecure_disable_ssl: bool = False

    x_log_level: int = 2**4  # logging.NOTSET
    x_internal_check_process: int = 1  # TODO: make configurable
    x_file_stream_retry_max: int = 2**2
    x_file_stream_retry_wait_min_seconds: float = 2 ** (-1)
    x_file_stream_retry_wait_max_seconds: float = 2
    x_file_stream_timeout_seconds: int = 2**5  # 2**2
    x_file_stream_max_conn: int = 2**5
    x_file_stream_max_size: int = 2**18
    x_file_stream_transmit_interval: int = 2**3
    x_sys_sampling_interval: int = 2**2
    x_sys_label: str = "_sys/"
    x_meta_label: str = "_meta/"

    url: str = "https://demo.mlop.ai"
    url_token: str = f"{url}/api-keys"
    _url_api: str = "https://api.mlop.ai"
    url_login: str = f"{_url_api}/api/is-logged-in"
    url_start: str = f"{_url_api}/api/create-run"
    url_stop: str = f"{_url_api}/api/update-run-status"
    url_meta: str = f"{_url_api}/api/add-logname"
    _url: str = "https://server.mlop.ai"
    url_data: str = f"{_url}/ingest/metrics"
    url_file: str = f"{_url}/files"
    url_message: str = f"{_url}/ingest/logs"
    url_view: str = f"{_url}/view"

    def update(self, settings) -> None:
        if isinstance(settings, Settings):
            settings = settings.to_dict()
        for key, value in settings.items():
            setattr(self, key, value)

    def to_dict(self) -> dict[str, any]:
        return {key: getattr(self, key) for key in self.__annotations__.keys()}

    def work_dir(self) -> str:
        return os.path.join(
            self.dir, "." + self.tag, self.project, self._op_name, self._op_id
        )

    def _nb(self) -> bool:
        return (
            get_console() in ["ipython", "jupyter"]
            or self._nb_colab()
            or self._nb_kaggle()
        )

    def _nb_colab(self) -> bool:
        return "google.colab" in sys.modules

    def _nb_kaggle(self) -> bool:
        return (
            os.getenv("KAGGLE_KERNEL_RUN_TYPE") is not None
            or "kaggle_environments" in sys.modules
            or "kaggle" in sys.modules
        )


class OpsSetup:
    def __init__(self, settings: Settings | None = None) -> None:
        self.settings = settings


def setup(settings: Settings | None = None) -> OpsSetup:
    logger.debug(f"{tag}: loading settings")
    return OpsSetup(settings=settings)


def get_console() -> str:
    try:
        from IPython import get_ipython

        ipython = get_ipython()
        if ipython is None:
            return "python"
    except ImportError:
        return "python"

    if "spyder" in sys.modules or "terminal" in ipython.__module__:
        return "ipython"

    connection_file = (
        ipython.config.get("IPKernelApp", {}).get("connection_file", "")
        or ipython.config.get("ColabKernelApp", {}).get("connection_file", "")
    ).lower()
    if "jupyter" not in connection_file:
        return "ipython"
    else:
        return "jupyter"



import builtins
import logging
import sys
import time

from .api import make_compat_message_v1
from .util import ANSI

logger = logging.getLogger(f"{__name__.split('.')[0]}")

bak_input = builtins.input
bak_stdout = sys.stdout
bak_stderr = sys.stderr

colors = {
    "DEBUG": ANSI.green,
    "INFO": ANSI.blue,
    "WARNING": ANSI.yellow,
    "ERROR": ANSI.red,
    "CRITICAL": ANSI.purple,
}
styles = {
    "DEBUG": " 💬 ",
    "INFO": " 🚀 ",
    "WARNING": " 🚨 ",
    "ERROR": " ⛔ ",
    "CRITICAL": " 🚫 ",
}


class ColorFormatter(logging.Formatter):
    def format(self, record):
        prefix = ANSI.bold + ANSI.blue + f"{__name__.split('.')[0]}:" + ANSI.reset
        color = colors.get(record.levelname, "")
        style = styles.get(record.levelname, "")
        # record.msg = f"{color}{record.msg}{ANSI.reset}"
        return f"{prefix}{color}{style}{super().format(record)}{ANSI.reset}"


class ConsoleHandler:
    def __init__(
        self, logger, queue, level=logging.INFO, stream=sys.stdout, type="stdout"
    ):
        self.logger = logger
        self.queue = queue
        self.level = level
        self.stream = stream
        self.type = type
        self.count = 0

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.count += 1
            m = line.rstrip()
            self.queue.put(
                make_compat_message_v1(self.level, m, int(time.time()), self.count)
            )
            self.logger.log(self.level, m)
        self.stream.write(buf)
        self.stream.flush()

    def flush(self):
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


def stream_formatter(settings):
    if settings.x_log_level <= logging.DEBUG:
        return ColorFormatter(
            "%(asctime)s.%(msecs)03d %(levelname)-8s %(threadName)-10s:%(process)d "
            "[%(filename)s:%(funcName)s():%(lineno)s] %(message)s",
            datefmt="%H:%M:%S",
        )
    else:
        return ColorFormatter(
            "%(asctime)s | %(message)s",
            # "%(asctime)s | %(levelname)-8s | %(message)s",
            datefmt="%H:%M:%S",
        )


def input_hook(prompt="", logger=None):
    content = bak_input(prompt)
    logger.warn(f"{prompt}{content}")
    return content


def setup_logger(settings, logger, console=None) -> None:
    if settings._nb_colab():
        rlogger = logging.getLogger()
        for h in rlogger.handlers[:]:  # iter root handlers
            rlogger.removeHandler(h)

    logger.setLevel(settings.x_log_level)

    if len(logger.handlers) == 0:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(stream_formatter(settings))
        logger.addHandler(stream_handler)

    if settings._op_id and not settings.disable_logger:
        if len(console.handlers) > 0:  # full logger
            return
        logger, console = setup_logger_file(settings, logger, console)


def teardown_logger(logger, console=None):
    for h in logger.handlers[:]:
        logger.removeHandler(h)
    if console:
        builtins.input = bak_input
        sys.stdout = bak_stdout  # global bak_stdout
        sys.stderr = bak_stderr
        teardown_logger(logger=console)


def setup_logger_file(settings, logger, console):
    console.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(f"{settings.work_dir()}/{settings.tag}.log")
    file_formatter = logging.Formatter(
        "%(asctime)s %(levelname)-8s %(threadName)-10s:%(process)d "
        "[%(filename)s:%(funcName)s():%(lineno)s] %(message)s"
    )
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    file_handler = logging.FileHandler(f"{settings.work_dir()}/sys.log")
    file_formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d | %(levelname)-7s | %(message)s",
        datefmt="%H:%M:%S",
    )
    file_handler.setFormatter(file_formatter)
    console.addHandler(file_handler)  # TODO: fix slow file writes

    if settings.mode == "debug":
        builtins.input = lambda prompt="": input_hook(prompt, logger=console)
    sys.stdout = ConsoleHandler(
        console, settings.message, logging.INFO, sys.stdout, "stdout"
    )
    sys.stderr = ConsoleHandler(
        console, settings.message, logging.ERROR, sys.stderr, "stderr"
    )

    return logger, console
